/*
	FEZ: Frame Evading Zoomrects
	
	 by Douglas McKenna
	 	(c) Copyright 1994 All rights reserved.

	 	Mathemaesthetics, Inc.
		PO Box 298  Boulder  CO  80306-0298 USA
		
		Submitted as a hack and a half for MacHack, June 24, 1994
	
	 	Permission is granted to use this code in any non-commercial product.
	 	Please contact the author for prior written approval for use in any
	 	commercial or shareware product.
	 	
	
	This file contains the shell application for creating window's with a
	double-clickable and dragable fez widget in them that opens into another
	window.  This provides a testbed for our FEZ routines, described in the
	accompanying file, "FEZ.c".
 */


#include "Main.h"
#include "FEZ.h"

/* Global variables for whole app */

SysEnvRec			thisMac;

/* Global variables for this file's routines only */

static EventRecord	myEvent;
static int			appleEventsOK,quitting;
static MenuHandle	appleMenu,fileMenu,demoMenu,optionsMenu,editMenu,windowMenu;
static CIconHandle	fezHat,fezHatSelected;

static BitMap		dragBits;
static GrafPort		dragPort;

static Rect			dragRect;
static WindowPtr	windows[MAXWINDOWS];
static Rect			windowPosition[MAXWINDOWS];
static Rect			windowWidgetPosition[MAXWINDOWS];
static short		whichDemo,lineThick;


/* Prototypes for local routines */

void			main(void);
void			Initialize(void);

void			InstallCoreEventHandlers(void);
void			DoHighLevelEvent(EventRecord *evt);
pascal OSErr	HandleOAPP(AppleEvent *appleEvent, AppleEvent *reply, long refcon);
pascal OSErr 	HandleODOC(AppleEvent *appleEvent, AppleEvent *reply, long refcon);
pascal OSErr 	HandlePDOC(AppleEvent *appleEvent, AppleEvent *reply, long refcon);
pascal OSErr 	HandleQUIT(AppleEvent *appleEvent, AppleEvent *reply, long refcon);
int				DoOpenApplication(void);
OSErr			MyGotRequiredParams(AppleEvent *appleEvent);

int				MainEvent(void);
int				DoMouseDown(void);
void			DoContentClick(WindowPtr aWindow, Point p);
void			SelectWidget(WidgetHandle item, int selected);
int				DoMouseUp(void);
void			DoActivate(WindowPtr w, int activ);
void			DoUpdate(WindowPtr w);
Rect *			SetInteriorClipping(WindowPtr w);
void			RestoreInteriorClipping(WindowPtr w);
int				DoKeyDown(void);
void			DoGrow(WindowPtr w);
int				DoMenu(long choice);

int				IsDoubleClick(Point *pt, long *time, int tol);
WindowPtr		AddWidgetWindow(Byte *title, Rect *pos, int stdPosition);
WindowPtr		MakeWidgetWindow(Byte *title, Rect *rect, Rect *initPos);
void			DoAppleMenu(int choice);
void			DispatchEvent(EventRecord *evt);
void			FixMenus(void);
void			CloseZoom(WindowPtr w);
void			CloseWidgetWindow(WindowPtr w);
void			OpenWidgetWindow(WindowPtr w, int stdPosition);
void			DrawWidget(WidgetHandle item);

void main()
	{
		Initialize();
		
		while (MainEvent())	;			/* Life until event-ual death */
	}


void Initialize()
	{
		long size; GrafPtr savePort;
		
		/* Usual Mac Toolbox startup calls */
		
		MaxApplZone();
		InitGraf(&qd.thePort);
		InitFonts();
		InitWindows();
		InitCursor();
		InitMenus();
		InitDialogs(0L);
		FlushEvents(everyEvent, 0 );
		
		/* Figure out what's legal on this machine */
		
		SysEnvirons(1,&thisMac);
		appleEventsOK = (thisMac.systemVersion >= 0x0700);	/* Should be using Gestalt */
		
		/* Install Apple Event stuff */
		
		if (appleEventsOK)
			InstallCoreEventHandlers();
		
		/* Install menus */
		
		appleMenu = GetMenu(appleMenuID);
		AddResMenu(appleMenu,'DRVR');
		InsertMenu(appleMenu,0);
		fileMenu = GetMenu(fileMenuID);
		InsertMenu(fileMenu,0);
		editMenu = GetMenu(editMenuID);
		InsertMenu(editMenu,0);
		demoMenu = GetMenu(demoMenuID);
		InsertMenu(demoMenu,0);
		optionsMenu = GetMenu(optionsMenuID);
		InsertMenu(optionsMenu,0);
		windowMenu = GetMenu(windowMenuID);
		InsertMenu(windowMenu,0);
		
		whichDemo = DM_Demo0;
		CheckItem(demoMenu,whichDemo,TRUE);
		lineThick = 1;
		
		DrawMenuBar();
		
		/* Precompute the widget image by building an offscreen 1-bit port and drawing into it */
		
		GetPort(&savePort);
		
			// Set up the bitmap
		
		SetRect(&dragBits.bounds,0,0,WIDGETWIDTH,WIDGETHEIGHT);
		dragBits.rowBytes = 2 * ((dragBits.bounds.right - dragBits.bounds.left) + 15) / 16;
		size = dragBits.rowBytes * (dragBits.bounds.bottom-dragBits.bounds.top);
		dragBits.baseAddr = (Ptr)NewPtrClear(size);
		
			// Create the offscreen port from which to copy widget image
		
		OpenPort(&dragPort);								// Calls SetPort()
		SetPortBits(&dragBits);
		RectRgn(dragPort.visRgn,&dragBits.bounds);
		ClipRect(&dragBits.bounds);
		
			// Draw the widget in the current offscreen port
			
		FillRect(&dragBits.bounds,&qd.gray);
		FrameRect(&dragBits.bounds);
		
			// Restore environment to whatever it was
			
		SetPort(savePort);
		
		/* Do pre-allocations */
		
		//clipRgn = NewRgn();
		//if (clipRgn == NIL) ExitToShell();
		//clipLevel = 0;
		
		fezHat = GetCIcon(fezHatColorIconID);
		if (fezHat == NIL) ExitToShell();
		fezHatSelected = GetCIcon(fezHatSelectedID);
		if (fezHatSelected == NIL) ExitToShell();
	}

/*
 *	Install the addresses of the functions that the AppleEvent manager will
 *	call when any of the standard core events come in over the transom.
 */

static void InstallCoreEventHandlers()
	{
		OSErr err;
		
		err = AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,HandleOAPP,0,FALSE);
		if (err) goto broken;
		err = AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,HandleODOC,0,FALSE);
		if (err) goto broken;
		err = AEInstallEventHandler(kCoreEventClass,kAEPrintDocuments,HandlePDOC,0,FALSE);
		if (err) goto broken;
		err = AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,HandleQUIT,0,FALSE);
broken:
		if (err) {
			}
	}

pascal OSErr HandleOAPP(AppleEvent *appleEvent, AppleEvent *reply, long refcon)
	{
		OSErr err = noErr, keepGoing = TRUE;
		
		err = MyGotRequiredParams(appleEvent);
		if (!err)
			keepGoing = DoOpenApplication();
		
		return(err);
	}

/*
 *	Check to see that all required parameters have been retrieved (IM 6-47)
 */

static OSErr MyGotRequiredParams(AppleEvent *appleEvent)
	{
		DescType returnedType;
		Size actualSize;
		OSErr err;
		
		err = AEGetAttributePtr(appleEvent,keyMissedKeywordAttr,typeWildCard,
															&returnedType,NIL,0,&actualSize);
		if (err == errAEDescNotFound) return(noErr);
		if (!err) err = errAEEventNotHandled;
		
		return(err);
	}
	
static int DoOpenApplication()
	{
		int keepGoing = TRUE;
				
		return(keepGoing);
	}

pascal OSErr HandleODOC(AppleEvent *appleEvent, AppleEvent *reply, long refcon)
	{
		OSErr err = noErr;
		long iFile,nFiles,numErrs=0;
		Size actualSize;
		AEKeyword keywd;
		DescType returnedType;
		AEDescList docList;
		FSSpec theFile;
		
		err = AEGetParamDesc(appleEvent, keyDirectObject, typeAEList, &docList);

		err = MyGotRequiredParams(appleEvent);
		if (!err) {
			AECountItems(&docList,&nFiles);
			for (iFile=1; iFile<=nFiles; iFile++) {
				err = AEGetNthPtr(&docList,iFile,typeFSS,&keywd,&returnedType,
												(Ptr)&theFile,sizeof(FSSpec),&actualSize);
				if (err) {
					break;
					}
				 else {
				 	WindowPtr w;
					w = AddWidgetWindow((Byte *)&theFile.name,NIL,TRUE);
					if (w) ShowWindow(w);
					}
				}
			}

		AEDisposeDesc(&docList);
		
		return(err);
	}

pascal OSErr HandlePDOC(AppleEvent *appleEvent, AppleEvent *reply, long refcon)
	{
		short err;
		
		err = HandleODOC(appleEvent,reply,refcon);
		return(err);
	}

pascal OSErr HandleQUIT(AppleEvent *appleEvent, AppleEvent *reply, long refcon)
	{
		int keepGoing = DoMenu(FM_Quit);
		quitting = !keepGoing;
		
		return(keepGoing);
	}


/*
 *	Process next event, return TRUE if okay to continue
 */

int MainEvent()
	{
		int gotOne;
		static int lastWasNull = FALSE;
		
		gotOne = WaitNextEvent(everyEvent,&myEvent,0,NIL);
		
    	if (gotOne) {
    		FixMenus();
    		DispatchEvent(&myEvent);
    		lastWasNull = FALSE;
    		}
    	 else {
    		/* Null event */
    		SetCursor(&qd.arrow);
    		if (!lastWasNull) FixMenus();
    		lastWasNull = TRUE;
			}
					
		return(!quitting);
	}

/*
 *	Figure out what kind of event we've gotten, and deal with it
 */

void DispatchEvent(EventRecord *evt)
	{
    	WindowPtr w; short result;
   	
   		quitting = FALSE;
   		
		switch (evt->what) {
			case mouseDown:
				quitting = DoMouseDown();
				break;
			case mouseUp:
				quitting = DoMouseUp();
				break;
			case keyDown:
			case autoKey:
				quitting = DoKeyDown();
				break;
			case activateEvt:
				w = (WindowPtr)evt->message;
				DoActivate(w,(myEvent.modifiers & activeFlag)!=0);
				break;
			case updateEvt:
				w = (WindowPtr)evt->message;
				DoUpdate(w);
				break;
				
			case diskEvt:
				result = HiWord(myEvent.message);
				if (result) {
					Point corner;
					/* ??? Can dialog box be centered	*/
					/*   There doesn't appear to be a	*/
					/*   DLOG resource for this one.	*/
					/*   Location (112,80) is the value	*/
					/*   used by Finder 6.1. 			*/
					SetPt(&corner, 112, 80);
					result = DIBadMount(corner, myEvent.message);
					}
				break;
        	    
			case app4Evt:
				switch(myEvent.message >> 24) {
					case mouseMovedMessage:
						break;
					case suspendResumeMessage:
						w = FrontWindow();
						if (w) 
							DoActivate(w,(myEvent.message & 1) != 0);
						break;
					}
				break;
				
			case kHighLevelEvent:
				DoHighLevelEvent(&myEvent);
				break;
			
			default:
				break;
			}
	}

static void DoHighLevelEvent(EventRecord *evt)
	{
		OSErr err = noErr;
		
		/* No high level events supported yet except for the core events */
		
		if (evt->message == (long)kCoreEventClass) {
			err = AEProcessAppleEvent(evt);
			}
	}

/*
 *	Call this for an activate/deactivate event
 */

void DoActivate(WindowPtr w, int activ)
	{
		SetPort(w);
		DrawGrowIcon(w);
	}

/*
 *	Call this to redraw contents of given window.  In this testbed, all windows
 *	are the same.
 */

void DoUpdate(WindowPtr w)
	{
		GrafPtr tmp; WidgetHandle item;
		
		GetPort(&tmp); SetPort(w);
		BeginUpdate(w);
		
		DrawGrowIcon(w);
		
		SetInteriorClipping(w);
		
		item = (WidgetHandle)GetWRefCon(w);
		DrawWidget(item);
		
		RestoreInteriorClipping(w);
		
		EndUpdate(w);
		SetPort(tmp);
	}

/*
 *	Clip the window to its content region within scroll bar bounds.  Delivers a pointer
 *	to the clipping rectangle.
 */

Rect *SetInteriorClipping(WindowPtr w)
	{
		static Rect clipper;
		
		clipper = w->portRect;
		clipper.right -= SCROLLBARWIDTH-1;
		clipper.bottom -= SCROLLBARWIDTH-1;
		ClipRect(&clipper);
		
		return(&clipper);
	}

/*
 *	Restore the current window's clipping region to whatever it was when
 *	when PushContent was called.
 */

void RestoreInteriorClipping(WindowPtr w)
	{
		ClipRect(&w->portRect);
	}

/*
 *	Draw the window's widget at its current position.  The handle to the window widget
 *	is always in the refCon field, and it's always defined.
 */
		
void DrawWidget(WidgetHandle item)
	{
		Rect position;
		
		position = (*item)->position;				// Local coordinates
		PlotCIcon(&position,(*item)->isSelected ? fezHatSelected : fezHat);
		// CopyBits(&dragBits,&qd.thePort->portBits,&dragBits.bounds,&position,srcCopy,NIL);
	}

/*
 *	Call this for a mouse up event
 */

int DoMouseUp()
	{
		return(FALSE);
	}

/*
 *	Call this for a mouse down click anywhere
 */

int DoMouseDown()
	{
		short part; Point p; int done = FALSE;
		WindowPtr aWindow;
		
		p = myEvent.where;
		part = FindWindow(p,&aWindow);
		if (aWindow == NIL)
			switch(part) {
				case inDesk:
						break;
				case inMenuBar:
						done = DoMenu(MenuSelect(p));
						break;
				}
		 else
			if (aWindow!=FrontWindow() && part!=inGrow)
				SelectWindow(aWindow);
			 else
				switch(part) {
					case inContent:
							DoContentClick(aWindow,p);
							break;
					case inGoAway:
							if (TrackGoAway(aWindow,p))
								CloseWidgetWindow(aWindow);
							break;
					case inZoomIn:
					case inZoomOut:
							if (TrackBox(aWindow,p,part)) {
								PushPort(aWindow);
								EraseRect(&aWindow->portRect);
								ZoomWindow(aWindow,part,TRUE);
								DrawGrowIcon(aWindow);
								InvalRect(SetInteriorClipping(aWindow));
								RestoreInteriorClipping(aWindow);
								PopPort();
								}
							break;
					case inSysWindow:
							SystemClick(&myEvent,aWindow);
							break;
					case inDrag:
							DragWindow(aWindow,p,&qd.screenBits.bounds);
							break;
					case inGrow:
							DoGrow(aWindow);
							break;
					}

		return(done);
	}	

/*
 *	Deal with a mouse down click in the window's content region at the global
 *	coordinate in given point.
 */

void DoContentClick(WindowPtr aWindow, Point p)
	{
		short dx,dy; long now; Rect rect,*clip; OSErr err;
		WidgetHandle item;
		GWorldPtr offscreen,saveWorld; PixMapHandle pm;
		GDHandle saveGDevice;
		
		// Get the widget in window, and its position
		
		item = (WidgetHandle)GetWRefCon(aWindow);
		rect = (*item)->position;
		GlobalToLocal(&p);
		
		// Don't overwrite scroll bar lines when dragging, selecting, or deselecting
		clip = SetInteriorClipping(aWindow);
		
		// If click is on widget, select it and entertain a drag or double-click
		// otherwise, deselect widget
		
		if (PtInRect(p,&rect)) {
			now = TickCount();
			if (IsDoubleClick(&p,&now,2))
				OpenWidgetWindow(aWindow,FALSE);
			 else {
				// First click always selects item if not already selected
				SelectWidget(item,TRUE);
				
				if (StillDown()) {
				
					// Prepare for a drag: use an offscreen graphics world to avoid flicker
					// If no offscreen world allocated, then we'll draw directly to screen
					// since both screen and offscreen will be using the same coordinates
					// Note that we could use temporary memory here, since we'll be throwing
					// it all away as soon as user lets up the mouse.
					
					GetGWorld(&saveWorld,&saveGDevice);
					err = NewGWorld(&offscreen,0,clip,NIL,NIL,noNewDevice);
					if (err == noErr) {
						SetGWorld(offscreen,NIL);
						pm = GetGWorldPixMap(offscreen);
						LockPixels(pm);
						EraseRect(clip);
						}
					 else
						SysBeep(1);
					
					HideCursor();					// So cursor doesn't flicker during drag
					
					while (WaitMouseUp()) {
						Point newPt;
						
						// Get new mouse position in window's local coordinates
						PushPort(aWindow);
						GetMouse(&newPt);
						PopPort();
						
						if (!EqualPt(p,newPt)) {
							// Mouse has moved to a new position: use offset to move widget
							dx = newPt.h - p.h;
							dy = newPt.v - p.v;
							EraseRect(&rect);			// Erase old icon bounds
							OffsetRect(&rect,dx,dy);	// Move icon to new position
							(*item)->position = rect;
							DrawWidget(item);			// Redraw it in current world
							// And update screen if above drawing was into offscreen world
							if (err == noErr)
								CopyBits((BitMap *)*pm,&aWindow->portBits,clip,clip,srcCopy,NIL);
							// Move to new point and wait until another change in mouse position
							p = newPt;
							}
						}
					ShowCursor();
					// Drag done: clean up
					if (err == noErr) {
						UnlockPixels(pm);
						DisposeGWorld(offscreen);
						}
					SetGWorld(saveWorld,saveGDevice);
					}
				}
			}
		 else {
			SelectWidget(item,FALSE);				// Deselect it
			}
		
		// Allow full content area to be drawn into again
		RestoreInteriorClipping(aWindow);
	}

/*
 *	Given a widget in a window, set its selection state to 0 or 1 (selected)
 *	and graphically draw the selection state now.
 */

void SelectWidget(WidgetHandle item, int selected)
	{
		if ((selected!=0) ^ ((*item)->isSelected!=0)) {		// Only if this is an actual change
			(*item)->isSelected = selected;
			SetInteriorClipping((*item)->itsWindow);
			DrawWidget(item);
			RestoreInteriorClipping((*item)->itsWindow);
			}
	}
	
void OpenWidgetWindow(WindowPtr parentWindow, int stdPosition)
	{
		Rect bounds,rect; Point pt;
		WindowPtr w; WidgetHandle item;
		ZoomRect start,end;
		
		if (parentWindow) {
			// Check first for window whose widget is already in this window
			item = (WidgetHandle)GetWRefCon(parentWindow);
			if ((*item)->openWindow)
				SelectWindow((*item)->openWindow);
			 else {
				SelectWidget(item,TRUE);		// For easy MacHack demo
				// Widget not yet opened: open a window for it
				w = AddWidgetWindow(NIL,&bounds, stdPosition);
				if (w) {
					(*item)->openWindow = w;
					rect = (*item)->position;
					pt.h = pt.v = 0;
					LocalToGlobal(&pt);
					OffsetRect(&rect,pt.h,pt.v);
					
					switch(whichDemo) {
						case DM_Demo0:
							break;
						case DM_Demo1:
							ZoomCenterRect(&rect,&bounds,lineThick);
							break;
						case DM_Demo2:
							// Prepare arguments to zoom routine
							start.win = parentWindow; start.zoom = rect;
							end.win = w; end.zoom = bounds;
							start.width = end.width = lineThick;
							// Do the zoom from start to end
							SingleSplineZoom(&start,&end,TRUE);
							break;
						case DM_Demo3:
							// Prepare arguments to zoom routine
							start.win = parentWindow; start.zoom = rect;
							end.win = w; end.zoom = bounds;
							start.width = end.width = lineThick;
							// Do the zoom from start to end
							SingleSplineZoom(&start,&end,TRUE);
							break;
						default:
							/*
							 *	If you know that the widget being opened is always in the
							 *	frontmost window, then you can always use LinearZoomRect()
							 *	because there will never be an intervening window between
							 *	the widget's window and the new one being opened.  However,
							 *	in general, you can't assume this, because there can be
							 *	floating windows open that do intervene and whose frames
							 *	the zooming rects should evade.
							 *
							 *	For now, we don't support floating windows.
							 */
							break;
						}
					
					ShowWindow(w);
					}
				}
			}
		 else {
			w = AddWidgetWindow(NIL,NIL,stdPosition);
			if (w) ShowWindow(w);
			}
	}

void CloseWidgetWindow(WindowPtr w)
	{
		short i; Point pt; WidgetHandle item;
		
		if (w) {
			// Find the entry in table, mark it free, and ditch the given window
			for (i=0; i<MAXWINDOWS; i++) {
				if (windows[i] == w) {
				
					// Save widget position (local coords) for re-opening
					item = (WidgetHandle)GetWRefCon(w);
					windowWidgetPosition[i] = (*item)->position;
		
					// Save global position of window on desktop for re-opening
					windowPosition[i] = w->portRect;
					PushPort(w);
					pt.h = pt.v = 0;
					LocalToGlobal(&pt);
					OffsetRect(&windowPosition[i],pt.h,pt.v);
					PopPort();	
		
					CloseZoom(w);
					DisposeWindow(w);
					windows[i] = NIL;
					break;
					}
				}
			}
	}

/*
 *	Determine if the given window was opened from a widget in some other window,
 *	and if so, zoom back to the widget position after hiding the window.
 */

void CloseZoom(WindowPtr w)
	{
		Rect startBox = (*((WindowPeek)w)->strucRgn)->rgnBBox;
		Rect endBox; WindowPeek wp,wBehind; Point pt;
		WidgetHandle item; ZoomRect start,end; ZoomRectHandle zoomArray;
		
		SetRect(&endBox,0,0,0,0);
		wp = (WindowPeek)FrontWindow();
		while (wp) {
			item = (WidgetHandle)GetWRefCon((WindowPtr)wp);
			// If not desk accessory and item window is this window
			if (wp->windowKind>=0 && item!=NIL && (*item)->openWindow==w) {
				// Detach item from its window, which is about to close
				(*item)->openWindow = NIL;
				endBox = (*item)->position;
				/* Keep track of widget position in global coordinates */
				SetPort((WindowPtr)wp);
				pt.h = pt.v = 0;
				LocalToGlobal(&pt);
				OffsetRect(&endBox,pt.h,pt.v);
				SetPort(w);
				break;
				}
			wp = wp->nextWindow;
			}
		
		item = (WidgetHandle)GetWRefCon(w);
		DisposeHandle((Handle)item);
		
		// Hide it since we still need it's port for certain zooms.  Before we muck
		// with the window order, though, precalculate the multi-window zoom information
		
		if (whichDemo == DM_Demo3) {
			start.zoom = startBox; start.win = w;
			end.zoom = endBox; end.win = (WindowPtr)wp;
			start.width = end.width = lineThick;
			zoomArray = GetMultiZoom(&start,&end,FALSE);
			}
		 else
			zoomArray = NIL;
		
		HideWindow(w);
		
		// Do the various flavors of zooming
		
		if (!EmptyRect(&endBox)) {
			DoUpdate((WindowPtr)wp);		// Make widget visible if it's behind closed window
			switch(whichDemo) {
				case DM_Demo0:
					break;
				case DM_Demo1:
					ZoomCenterRect(&startBox,&endBox,lineThick);
					break;
				case DM_Demo2:
					start.zoom = startBox; start.win = w;
					end.zoom = endBox; end.win = (WindowPtr)wp;
					start.width = end.width = lineThick;
					SingleSplineZoom(&start,&end,FALSE);
					break;
				case DM_Demo3:
					MultipleSplineZoom(zoomArray,FALSE);
					break;
				default:
					break;
				}
			}
	}

/*
 *	Called for a key down event
 */

int IsDoubleClick(Point *pt, long *time, int tol)
	{
		int ans = FALSE;
		static long lastTime; static Point thePt;
		
		if (*time-lastTime < GetDblTime()) {
			Rect box;
			SetRect(&box,thePt.h-tol,thePt.v-tol,thePt.h+tol,thePt.v+tol);
			ans = PtInRect(*pt,&box);
			lastTime = 0;
			thePt.h = thePt.v = -16000;
			}
		 else {
			lastTime = *time;
			thePt = *pt;
			}
		return(ans);
	}

int DoKeyDown()
	{
		int ch = (unsigned)(myEvent.message & charCodeMask);
		
		if (myEvent.modifiers & cmdKey) {
			return(DoMenu(MenuKey(ch)));
			}

		return(FALSE);
	}

int DoMenu(long choice)
	{
		short menuID = HiWord(choice);
		short menuCh = LoWord(choice);
		short i,quitting = FALSE;
		short newThick;
		WindowPtr w;
		
		switch(menuID) {
			case appleMenuID:
				DoAppleMenu(menuCh);
				break;
			case fileMenuID:
				switch(menuCh) {
					case FM_New:
						OpenWidgetWindow(FrontWindow(),TRUE);
						break;
					case FM_Open:
						OpenWidgetWindow(FrontWindow(),FALSE);
						break;
					case FM_Close:
						CloseWidgetWindow(FrontWindow());
						break;
					case FM_Quit:
						quitting = TRUE;
						break;
					}
				break;
			case editMenuID:
				switch(menuCh) {
					case EM_SelectAll:
						w = FrontWindow();
						if (w) {
							WidgetHandle item = (WidgetHandle)GetWRefCon(w);
							SelectWidget(item,TRUE);
							}
						break;
					}
				break;
			case demoMenuID:
				CheckItem(demoMenu,whichDemo,FALSE);
				CheckItem(demoMenu,whichDemo = menuCh,TRUE);
				break;
			case optionsMenuID:
				newThick = GetLineThick(lineThick);
				if (newThick) lineThick = newThick;
				break;
			case windowMenuID:
				for (i=0; i<MAXWINDOWS; i++)
					if (windows[i] && --menuCh==0) {
						SelectWindow(windows[i]);
						break;
						}
				break;
			}
		
		HiliteMenu(0);
		return(quitting);
	}

static void FixMenus()
	{
		WindowPtr w = FrontWindow();
		short i,n; WidgetHandle item;
		
		if (w) {
			item = (WidgetHandle)GetWRefCon(w);
			if ((*item)->isSelected)
				EnableItem(fileMenu,FM_Open);
			 else
				//DisableItem(fileMenu,FM_Open);
				EnableItem(fileMenu,FM_Open);		// Makes MacHack demo faster
			}
		 else
			DisableItem(fileMenu,FM_Open);
		
		if (w) EnableItem(fileMenu,FM_Close);
		 else  DisableItem(fileMenu,FM_Close);
		
		for (i=EM_Undo; i<=EM_Paste; i++)
			DisableItem(editMenu,i);
		if (w) EnableItem(editMenu,EM_SelectAll);
		 else  DisableItem(editMenu,EM_SelectAll);
		if (w) EnableItem(editMenu,0);
		 else  DisableItem(editMenu,0);
		 
		// Rebuild the window menu
		
		n = CountMItems(windowMenu);
		while (n > 0)
			DeleteMenuItem(windowMenu,n--);
		
		if (w) EnableItem(windowMenu,0);
		 else	DisableItem(windowMenu,0);
		
		for (i=0; i<MAXWINDOWS; i++)
			if (windows[i]) {
				WindowPeek wp = (WindowPeek)windows[i];
				HLock((Handle)wp->titleHandle);
				AppendMenu(windowMenu,*wp->titleHandle);
				HUnlock((Handle)wp->titleHandle);
				}
		DrawMenuBar();
	}

/*
 *	Deal with a user choice from the Apple Menu: if not choosing "About Resorcerer..."
 *	then it's a Desk Accessory to be opened.
 */

static void DoAppleMenu(int choice)
	{
		Byte accName[256];
		
		switch(choice) {
			case AM_About:
				Alert(aboutBoxAlertID,NIL);
				break;
			default:
				GetItem(appleMenu,choice,accName);
				OpenDeskAcc(accName);
				break;
			}
	}

/*
 *	Grow a window to a new size in response to click in grow box
 */

void DoGrow(WindowPtr window)
	{
		Rect r;
		long newSize; int x,y;
		
		r = qd.screenBits.bounds;
		SetRect(&r,32,32,r.right-r.left,r.bottom-r.top);
		newSize = GrowWindow(window,myEvent.where,&r);
		if (newSize) {
			x = LoWord(newSize); y = HiWord(newSize);
			SizeWindow(window,x,y,TRUE);
			PushPort(window);
			EraseRect(&window->portRect);
			DrawGrowIcon(window);
			r = *SetInteriorClipping(window);
			InvalRect(&r);
			RestoreInteriorClipping(window);
			PopPort();
			}
	}

/*
 *	Create another window and its widget, and keep track of window by keeping
 *	its pointer in the first free slot in the table of pointers.  The initial
 *	position and title of the window are expected as arguments.  The window's
 *	refCon field is used to hold a Handle to the window's widget information.
 *	Returns NIL if not enough memory or other problem.
 */

WindowPtr MakeWidgetWindow(Byte *title, Rect *rect, Rect *initWidgetPos)
	{
		WindowPtr w=NIL; WidgetHandle item;
		
		item = (WidgetHandle)NewHandleClear(sizeof(Widget));
		if (item) {
			(*item)->position = *initWidgetPos;
			(*item)->isSelected = FALSE;
			
			w = NewCWindow(NIL,rect,title,FALSE,zoomDocProc,(WindowPtr)-1,TRUE,(long)item);
			if (w == NIL)
				DisposeHandle((Handle)item);
			 else
				(*item)->itsWindow = w;
			}
		
		return(w);
	}

/*
 *	Allocate another window on desktop, staggered, with given title,
 *	or (if title is NIL) with default title based on window index.
 *	Delivers window pointer, or NIL if unsuccessful.  If successful,
 *	sets *pos to the rectangle used to position the window (its
 *	content bounds).  Window remains hidden, and the current port
 *	remains the same.
 */

WindowPtr AddWidgetWindow(Byte *title, Rect *pos, int stdPosition)
	{
		short i,x,y; Rect rect,oldPos; Point pt; Byte *str;
		static Rect initPos = { 50, 50, 50+WIDGETHEIGHT, 50+WIDGETWIDTH };
		
		// Allocate first free slot from window table
		
		for (i=0; i<MAXWINDOWS; i++)
			if (windows[i] == NIL) {
			
				if (EmptyRect(&windowPosition[i]) || pos==NIL || stdPosition) {
					// Stagger window absolutely on main screen
					x = i / 8;
					y = i % 8;
					pt.h = 4;
					pt.v = 4 + GetMBarHeight();
					pt.h += 5 + 30*x + 5*y;
					pt.v += 20 + 20*y + 5*x;
					SetRect(&rect,100,100,300,300);
					OffsetRect(&rect,pt.h-rect.left,pt.v-rect.top);
					oldPos = initPos;
					}
				 else {
					// Use previous position
					rect = windowPosition[i];
					oldPos = windowWidgetPosition[i];
					}
				
				// Tell caller if caller wants answer
				if (pos) {
					*pos = rect;
					// This really should be the bounds of strucRgn, which is undefined
					// right now because the window hasn't been shown yet.
					pos->top -= TITLEBARHEIGHT;
					}
				// Give it its title using index in table
				if (title == NIL) {
					str = "\pFez xx";
					if (i < 10)
						{ str[5] = '0'+i; str[0] = 5; }
					 else
						{ str[5] = '0'+i/10; str[6] = '0'+i%10; str[0] = 6; }
					}
				 else
					str = title;
				
				// Get the window with widget already installed
				windows[i] = MakeWidgetWindow(str,&rect,&oldPos);
				return(windows[i]);
				}

		return(NIL);
	}

